# Copyright (c) 2024 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)


# Pyrat-Bay version:
TSO_VER = 0  # Major version
TSO_MIN = 4  # Minor version
TSO_REV = '0'  # Revision


__version__ = f'{TSO_VER}.{TSO_MIN}.{TSO_REV}'

